unit FastDIB;

interface

uses Windows;

const
{$IFDEF VER90}
  hSection=nil;
{$ELSE}
  hSection=0;
{$ENDIF}

type
TFColor  = record b,g,r:Byte; end;
TFColorA = record b,g,r,a:Byte; end;
PFColor  =^TFColor;
PFColorA =^TFColorA;
TFColorTable = array[0..255]of TFColorA;
PFColorTable =^TFColorTable;

TLines    = array[0..$FFFF]of Pointer;  PLines    =^TLines;
TLine8    = array[0..$FFFF]of Byte;     PLine8    =^TLine8;
TLine16   = array[0..$FFFF]of Word;     PLine16   =^TLine16;
TLine24   = array[0..$FFFF]of TFColor;  PLine24   =^TLine24;
TLine32   = array[0..$FFFF]of TFColorA; PLine32   =^TLine32;
TPixels8  = array[0..$FFFF]of PLine8;   PPixels8  =^TPixels8;
TPixels16 = array[0..$FFFF]of PLine16;  PPixels16 =^TPixels16;
TPixels24 = array[0..$FFFF]of PLine24;  PPixels24 =^TPixels24;
TPixels32 = array[0..$FFFF]of PLine32;  PPixels32 =^TPixels32;

TBMInfo = record
  bmiHeader: TBitmapInfoHeader;
  case Boolean of
  True:  (bmiColors: array[0..255]of TFColorA);
  False: (r,g,b:Longint);
end;

TFastDIB = class
  hDC,
  BWidth,
  Height,
  Gap,
  Mask,
  Handle:     Integer;
  Bits:       PLine8;
  Colors:     PFColorTable;
  bmInfo:     TBMInfo;
  Scanlines:  PLines;
  bshr,
  gshr,gshl,
  rshr,rshl:  Byte;
  Pixels8:    PPixels8;
  Pixels16:   PPixels16;
  Pixels24:   PPixels24;
  Pixels32:   PPixels32;

  constructor Create;
  destructor  Destroy; override;

  property    Compression:cardinal read bmInfo.bmiHeader.biCompression write bmInfo.bmiHeader.biCompression;
  property    BHeight:Longint read bmInfo.bmiHeader.biHeight write bmInfo.bmiHeader.biHeight;
  property    Width:Longint read bmInfo.bmiHeader.biWidth write bmInfo.bmiHeader.biWidth;
  property    Size:cardinal read bmInfo.bmiHeader.biSizeImage write bmInfo.bmiHeader.biSizeImage;
  property    Bpp:Word read bmInfo.bmiHeader.biBitCount write bmInfo.bmiHeader.biBitCount;
  property    rMask:Longint read bmInfo.r write bmInfo.r;
  property    gMask:Longint read bmInfo.g write bmInfo.g;
  property    bMask:Longint read bmInfo.b write bmInfo.b;

  function    GetPixel(y,x:Integer):TFColor;
  function    GetPixelB(y,x:Integer):Byte;
  procedure   SetPixel(y,x:Integer;p:TFColor);
  procedure   SetPixelB(y,x:Integer;p:Byte);
  property    Pixels[y,x:Integer]:TFColor read GetPixel write SetPixel;
  property    PixelsB[y,x:Integer]:Byte read GetPixelB write SetPixelB;

  procedure   SetSize(cx,cy:Integer;fBpp:Byte;fMask:Integer);
  procedure   LoadFromHandle(hBmp:Integer;SaveBpp:Boolean);
  procedure   LoadFromFile(sfile:string;SaveBpp:Boolean);
  procedure   SaveToFile(FileName:string);
  procedure   LoadFrom24_and_DitherTo16or15(FB:TFastDIB;fMask:Integer);
  procedure   Draw(fdc,x,y:Integer);
  procedure   Stretch(fdc,x,y,w,h:Integer);
  procedure   DrawRect(fdc,x,y,w,h,sx,sy:Integer);
  procedure   StretchRect(fdc,x,y,w,h,sx,sy,sw,sh:Integer);

  //conversions
  procedure   Convert16to16(FB:TFastDIB);
  procedure   Convert16to24(FB:TFastDIB);
  procedure   Convert24to16(FB:TFastDIB);
end;

function Get16Mask:Integer;

implementation

function Get16Mask:Integer;
var
sDC,bDC,hBM,c: Integer;
begin
  sDC:=GetDC(0);
  bDC:=CreateCompatibleDC(sDC);
  hBM:=CreateCompatibleBitmap(sDC,1,1);
  SelectObject(bDC,hBM);
  SetPixel(bDC,0,0,RGB(0,100,0));
  c:=GetPixel(bDC,0,0);
  DeleteDC(bDC);
  DeleteObject(hBM);
  ReleaseDC(0,sDC);
  if GetGValue(c)=100 then Result:=565 else Result:=555;
end;

constructor TFastDIB.Create;
begin
  FillChar(bmInfo,SizeOf(bmInfo),0);
  Colors:=@bmInfo.bmiColors;
  bmInfo.bmiHeader.biSize:=SizeOf(TBitmapInfoHeader);
  bmInfo.bmiHeader.biPlanes:=1;
  hDC:=CreateCompatibleDC(0);
  BWidth:=0; Height:=0; Gap:=0; Mask:=0;
  Handle:=0; bshr:=0; gshr:=0; gshl:=0; rshr:=0; rshl:=0;
end;

destructor TFastDIB.Destroy;
begin
  ReallocMem(Scanlines,0);
  DeleteDC(hDC);
  DeleteObject(Handle);
  inherited Destroy;
end;

procedure TFastDIB.SetSize(cx,cy:Integer;fBpp:Byte;fMask:Integer);
var
Base:    Longint;
sdc,i,x: Integer;
p:       Pointer;
n,b:     Byte;
begin
  if(fBpp=Bpp)and(Width=cx)and(Height=cy)and(fMask=Mask)then Exit;
  if fBpp=0 then
  begin
    sdc:=GetDC(0);
    Bpp:=GetDeviceCaps(sdc,BITSPIXEL);
    ReleaseDC(0,sdc);
  end else Bpp:=fBpp;
  if fMask=0 then if Bpp=16 then fMask:=Get16Mask else fMask:=888;
  Width:=cx; Height:=Abs(cy); BHeight:=-cy; Mask:=fMask;
  BWidth:=((Width*Bpp+31)and -32)shr 3; Size:=BWidth*Height;
  Gap:=BWidth-(Width*(Bpp div 8));

  if(Bpp=16)or(Bpp=32)then
  begin
    Compression:=BI_BITFIELDS; //parse mask & set shifts
    if Bpp=16 then Base:=$FFFF else Base:=$FFFFF; n:=0;
    b:=fMask mod 10; Inc(n,b); fMask:=fMask div 10;
    bMask:=Base shr(Bpp-n); bshr:=8-b; gshl:=b;
    b:=fMask mod 10; Inc(n,b); fMask:=fMask div 10;
    gMask:=Base shr(Bpp-n)and not bMask; rshl:=n; gshr:=8-b;
    b:=fMask mod 10; Inc(n,b); fMask:=fMask div 10;
    rMask:=Base shr(Bpp-n)and not(bMask or gMask); rshr:=8-b;
  end else Compression:=BI_RGB;

  DeleteObject(Handle);
  Handle:=CreateDIBSection(0,PBitmapInfo(@bmInfo)^,0,Pointer(Bits),hSection,0);
  SelectObject(hDC,Handle);
  ReallocMem(Scanlines,Height*4);
  x:=Integer(Bits);
  for i:=0 to Height-1 do
  begin
    Scanlines[i]:=Ptr(x);
    Inc(x,BWidth);
  end;
  Pixels8:=Pointer(Scanlines);
  Pixels16:=Pointer(Scanlines);
  Pixels24:=Pointer(Scanlines);
  Pixels32:=Pointer(Scanlines);
end;

procedure TFastDIB.SetPixel(y,x:Integer;p:TFColor);
begin
  case Bpp of
    16: Pixels16[y,x]:=
          p.r shr rshr shl rshl or
          p.g shr gshr shl gshl or
          p.b shr bshr;
    24: Pixels24[y,x]:=p;
    32: PDWord(@Pixels32[y,x])^:=
          p.r shr rshr shl rshl or
          p.g shr gshr shl gshl or
          p.b shr bshr;
  end;
end;

function TFastDIB.GetPixel(y,x:Integer):TFColor;
var
p:  Word;
pd: DWord;
begin
p := 0;
  case Bpp of
    16: begin
          p:=Pixels16[y,x];
          Result.b:=p shl bshr;
          Result.g:=p shr gshl shl gshr;
          Result.r:=p shr rshl shl rshr;
        end;
    24: Result:=Pixels24[y,x];
    32: begin
          pd:=PDWord(@Pixels32[y,x])^;
          Result.b:=p shl bshr;
          Result.g:=p shr gshl shl gshr;
          Result.r:=p shr rshl shl rshr;
        end;
  end;
end;

procedure TFastDIB.SetPixelB(y,x:Integer;p:Byte);
var
pb: PByte;
begin
  case Bpp of
    1: begin
         pb:=@Pixels8[y,x shr 3];
         pb^:=pb^ or p shl(7-(x mod 8));
       end;
    4: begin
         pb:=@Pixels8[y,x shr 1];
         if(x and 1)=0 then pb^:=pb^ or p shl 4
         else pb^:=pb^ or p;
       end;
    8: Pixels8[y,x]:=p;
  end;
end;

function TFastDIB.GetPixelB(y,x:Integer):Byte;
var
b: Byte;
begin
  case Bpp of
    1: begin
         b:=7-(x mod 8);
         Result:=Pixels8[y,x shr 3]and(1 shl b)shr b;
       end;
    4: if(x and 1)=0 then Result:=Pixels8[y,x shr 1]shr 4 and $F
       else Result:=Pixels8[y,x shr 1]and $F;
    8: Result:=Pixels8[y,x];
    else
    result := 0;
  end;

end;

procedure TFastDIB.LoadFromHandle(hBmp:Integer;SaveBpp:Boolean);
var
Bmp:   TBitmap;
memDC: Integer;
begin
  GetObject(hBmp,SizeOf(Bmp),@Bmp);
  if SaveBpp then SetSize(Bmp.bmWidth,Bmp.bmHeight,Bpp,Mask)
  else SetSize(Bmp.bmWidth,Bmp.bmHeight,Bmp.bmBitsPixel,Mask);
  memDC:=CreateCompatibleDC(0);
  SelectObject(memDC,hBmp);
  GetDIBits(memDC,hBmp,0,Height,Bits,PBitmapInfo(@bmInfo)^,0);
  DeleteDC(memDC);
end;

procedure TFastDIB.LoadFromFile(sfile:string;SaveBpp:Boolean);
var
hBmp: Integer;
begin
  hBmp:=LoadImage(0,PChar(sfile),IMAGE_BITMAP,0,0,LR_LOADFROMFILE or LR_CREATEDIBSECTION);
  LoadFromHandle(hBmp,SaveBpp);
  DeleteObject(hBmp);
end;

procedure TFastDIB.SaveToFile(FileName:string);
var
cSize,
hFile:  Integer;
iWrote: {$IFDEF VER90}Integer{$ELSE}Cardinal{$ENDIF};
fHead:  TBitmapFileHeader;
begin
csize := 0;
  case Bpp of
    1: cSize:=8;
    4: cSize:=64;
    8: cSize:=1024;
    16,24,32: cSize:=0;
  end;
  hFile:=CreateFile(PChar(FileName),GENERIC_WRITE,0,nil,CREATE_ALWAYS,FILE_FLAG_SEQUENTIAL_SCAN,0);
  fHead.bfType:=$4D42; //"BM"
  fHead.bfSize:=54+Size+cardinal(cSize);
  fHead.bfOffBits:=54+cSize;
  WriteFile(hFile,fHead,14,iWrote,nil);
  BHeight:=Abs(BHeight);
  WriteFile(hFile,bmInfo,40+cSize,iWrote,nil);
  BHeight:=-BHeight;
  WriteFile(hFile,Bits^,Size,iWrote,nil);
  CloseHandle(hFile);
end;

procedure TFastDIB.Draw(fdc,x,y:Integer);
begin
  StretchDIBits(fdc,x,y,Width,Height,0,0,Width,Height,
    Bits,PBitmapInfo(@bmInfo)^,0,SRCCOPY);
end;

procedure TFastDIB.Stretch(fdc,x,y,w,h:Integer);
begin
  SetStretchBltMode(fdc,STRETCH_DELETESCANS);
  StretchDIBits(fdc,x,y,w,h,0,0,Width,Height,Bits,
    PBitmapInfo(@bmInfo)^,0,SRCCOPY);
end;

procedure TFastDIB.DrawRect(fdc,x,y,w,h,sx,sy:Integer);
begin
  StretchDIBits(fdc,x,y,w,h,sx,sy,w,h,Bits,
    PBitmapInfo(@bmInfo)^,0,SRCCOPY);
end;

procedure TFastDIB.StretchRect(fdc,x,y,w,h,sx,sy,sw,sh:Integer);
begin
  SetStretchBltMode(fdc,STRETCH_DELETESCANS);
  StretchDIBits(fdc,x,y,w,h,sx,sy,sw,sh,Bits,
    PBitmapInfo(@bmInfo)^,0,SRCCOPY);
end;

procedure TFastDIB.Convert16to16(FB:TFastDIB);
var
x,y:     Integer;
pw1,pw2: PWord;
r,g,b:   Byte;
begin
  if Mask=FB.Mask then Move(FB.Bits^,Bits^,Size) else
  begin
    pw1:=Pointer(Bits);
    pw2:=Pointer(FB.Bits);
    for y:=0 to Height-1 do
    begin
      for x:=0 to Width-1 do
      begin
        b:=pw2^ shl FB.bshr;
        g:=pw2^ shr FB.gshl shl FB.gshr;
        r:=pw2^ shr FB.rshl shl FB.rshr;
        pw1^:=r shr rshr shl rshl or g shr gshr shl gshl or b shr bshr;
        Inc(pw1); Inc(pw2);
      end;
      pw1:=Ptr(Integer(pw1)+Gap);
      pw2:=Ptr(Integer(pw2)+FB.Gap);
    end;
  end;
end;

procedure TFastDIB.Convert24to16(FB:TFastDIB);
var
x,y: Integer;
pw:  PWord;
pc:  PFColor;
begin
  pw:=Pointer(Bits);
  pc:=Pointer(FB.Bits);
  for y:=0 to Height-1 do
  begin
    for x:=0 to Width-1 do
    begin
      pw^:=pc.r shr rshr shl rshl or pc.g shr gshr shl gshl or pc.b shr bshr;
      Inc(pw); Inc(pc);
    end;
    pw:=Ptr(Integer(pw)+Gap);
    pc:=Ptr(Integer(pc)+FB.Gap);
  end;
end;

procedure TFastDIB.Convert16to24(FB:TFastDIB);
var
x,y: Integer;
pw:  PWord;
pc:  PFColor;
begin
  pw:=Pointer(FB.Bits);
  pc:=Pointer(Bits);
  for y:=0 to Height-1 do
  begin
    for x:=0 to Width-1 do
    begin
      pc.b:=pw^ shl FB.bshr;
      pc.g:=pw^ shr FB.gshl shl FB.gshr;
      pc.r:=pw^ shr FB.rshl shl FB.rshr;
      Inc(pw); Inc(pc);
    end;
    pw:=Ptr(Integer(pw)+FB.Gap);
    pc:=Ptr(Integer(pc)+Gap);
  end;
end;




procedure TFastDIB.LoadFrom24_and_DitherTo16or15 (FB:TFastDIB;fMask:Integer);
type
TThreeInt   = record r,g,b:Integer;end;
PThreeInt   =^TThreeInt;
TLineErrors = array [-1..-1] of TThreeInt;
PLineErrors =^TLineErrors;

var
n,A,B,C:Integer;
pc,
Cl:      PFColor;
PW:      PWord;
Lines:   array[0..1]of PLineErrors;
Ln,Nxt:  Integer;
PTI:     PThreeInt;
Dir:     ShortInt;
_gshr,_bshr,_rshr,
_gshl,_rshl: Byte;
_r,_g,_b: Byte;
RNew,GNew,BNew:Byte;
rmsk,rtrs,rmax:Byte;
gmsk,gtrs,gmax:Byte;
bmsk,btrs,bmax:Byte;
begin
  SetSize(FB.Width,FB.Height,24,0);
  n:=0;
  _b:=fMask mod 10; Inc(n,_b); fMask:=fMask div 10; _bshr:=8-_b; _gshl:=_b;
  _g:=fMask mod 10; Inc(n,_g); fMask:=fMask div 10; _rshl:=n;   _gshr:=8-_g;
  _r:=fMask mod 10; Inc(n,_r); fMask:=fMask div 10; _rshr:=8-_r;
  rmax:=(1 shl _r)-1; rtrs:=3; rmsk:=7;
  gmax:=(1 shl _g)-1; gtrs:=3; gmsk:=7; //trs:=1 msk:=3
  bmax:=(1 shl _b)-1; btrs:=3; bmsk:=7;

  GetMem(Lines[0],24*(FB.Width+2));
  GetMem(Lines[1],24*(FB.Width+2));

  Cl:=PFColor(FB.Bits);
  for B:=0 to FB.Width-1 do
  begin
    Lines[0,B].r:=Cl.r*16;
    Lines[0,B].g:=Cl.g*16;
    Lines[0,B].b:=Cl.b*16;
    Inc(Cl);
  end;
  Cl:=Pointer(Integer(Cl)+FB.Gap);

  Dir:=1;
  for A:=1 to FB.Height do
  begin
    Nxt:=A mod 2;
    Ln:=1-Nxt;
    if A<FB.Height then
    begin
      for B:=0 to FB.Width-1 do
      begin
        Lines[Nxt,B].r:=Cl.r*16;
        Lines[Nxt,B].g:=Cl.g*16;
        Lines[Nxt,B].b:=Cl.b*16;
        Inc(Cl);
      end;
      Cl:=Pointer(Integer(Cl)+FB.Gap);
    end;

    B:=0;
    if Dir=-1 then B:=FB.Width-1;
    PTI:=@Lines[Ln,B];
    pc:=@Pixels24[A-1,B];
    while((B>-1)and(B<FB.Width))do
    begin
      PTI.r:=PTI.r div 16; PTI.g:=PTI.g div 16; PTI.b:=PTI.b div 16;
      if PTI.r>255 then PTI.r:=255 else if PTI.r<0 then PTI.r:=0;
      if PTI.g>255 then PTI.g:=255 else if PTI.g<0 then PTI.g:=0;
      if PTI.b>255 then PTI.b:=255 else if PTI.b<0 then PTI.b:=0;

      RNew:=PTI.r shr _rshr; GNew:=PTI.g shr _gshr; BNew:=PTI.b shr _bshr;
      if(((PTI.r and rmsk)>rtrs)and(RNew<rmax))then Inc(RNew);
      if(((PTI.g and gmsk)>gtrs)and(GNew<gmax))then Inc(GNew);
      if(((PTI.b and bmsk)>btrs)and(BNew<bmax))then Inc(BNew);

      //PW^:=BNew or GNew shl gshl or RNew shl rshl;
      pc.r:=RNew shl _rshr;
      pc.g:=GNew shl _gshr;
      pc.b:=BNew shl _bshr;

      C:=PTI.r - (RNew shl _rshr);
      if C<>0 then
      begin
        Inc(Lines[Ln,B+Dir].r,C*7);
        Inc(Lines[Nxt,B-Dir].r,C*3);
        Inc(Lines[Nxt,B].r,C*5);
        Inc(Lines[Nxt,B+Dir].r,C);
      end;
      C:=PTI.g - (GNew shl _gshr);
      if C<>0 then
      begin
        Inc(Lines[Ln,B+Dir].g,C*7);
        Inc(Lines[Nxt,B-Dir].g,C*3);
        Inc(Lines[Nxt,B].g,C*5);
        Inc(Lines[Nxt,B+Dir].g,C);
      end;
      C:=PTI.b - (BNew shl _bshr);
      if C<>0 then
      begin
        Inc(Lines[Ln,B+Dir].b,C*7);
        Inc(Lines[Nxt,B-Dir].b,C*3);
        Inc(Lines[Nxt,B].b,C*5);
        Inc(Lines[Nxt,B+Dir].b,C);
      end;
      Inc(pc,Dir);
      Inc(PTI,Dir);
      Inc(B,Dir);
    end;
    Dir:=-Dir;
  end;
  Dispose(Lines[0]);
  Dispose(Lines[1]);
end;

end.
